# -*- coding: utf-8 -*-
"""
Created on Wed Oct 27 15:17:35 2021

@author: Alexandra Klemme
"""

#%%% Import Libraries
import numpy as np
import pandas as pd
from scipy.optimize import curve_fit
from scipy import stats

#%%% Import Data
Datafile = './Data/DataTable.xlsx'                                          ### Location of supplementary data file                                          
List = pd.read_excel(Datafile, index_col=0)                   
Data = pd.DataFrame(List) 
Data = Data.drop('Simunjan_high')                                           ### Exclusion of Simunjan_2 campaigns

#%%% Prepare Data for optimization
mask = ~np.ma.masked_invalid(Data['CO2(uM)']).mask * ~np.ma.masked_invalid(Data['O2(uM)']).mask

X_vals = np.array([ np.concatenate(( Data['DOC(uM)'][mask], Data['DOC(uM)'][mask] )), 
                    np.concatenate(( Data['O2(uM)'][mask],  Data['O2(uM)'][mask]  )), 
                    np.concatenate(( Data['pH'][mask],      Data['pH'][mask]      )), 
                    np.concatenate(( Data['T(C)'][mask],    Data['T(C)'][mask]    )) ])

#%%% Define RMSE function
def rmse(predictions, targets):
    return np.sqrt(((predictions - targets) ** 2).mean())

#%%% Set form of decomposition approach

decomposition_form = 'exp'

### possible forms of decomposition:
### 1) 'exp' = exponential pH limitation and Michaelis-Menten O2 limitation (Delta_CO2 = R * DOC * O2/(O2+Km) * exp(L*(pH-pH_0)))
### 2) 'lin' = linear pH limitation and Michaelis-Menten O2 limitation (Deta_CO2 = R * DOC * O2/(O2+Km) * pH/pH_0)
### 3) 'exp_pH0' = exponetial pH limitation with variable pH_0 and Miachaelis-Menten O2 limitation (Delta_CO2 = R * DOC * O2/(O2+Km) * exp(L*(pH-pH_0)))
### 4) 'pH_only' = exponential pH limitation and no O2 limitation (Delta_CO2 = R * DOC * exp(L * (pH-pH_0)))
### 5) 'O2_only' = no pH limitation and Michaelis-Menten O2 limitation (Delta_CO2 = R * DOC * O2/(O2+Km))

#%%% Define functions to be optimized
if (decomposition_form == 'exp'):
    def function(x, a, b, c, d):
        
        Len = int(len(x[0])/2)
        DOC = x[0][:Len] ; O2 = x[1][:Len] ; pH = x[2][:Len]  ; T = x[3][:Len]                                    ## umol/L ; umol/L ; - ; °C
        
        k600_depth = 0.250751964                                                                                  ## cm/h/dm  
        
        O2_air   = 241.                                                                                           ## umol/L
        pCO2_air = 400.                                                                                           ## uatm
        Sc_O2  = 1568 - 86.04 * T + 2.142 * T**2 - 0.0216 * T**3
        Sc_CO2 = 1742 - 91.24 * T + 2.208 * T**2 - 0.0219 * T**3
        k_O2_depth  = k600_depth * (Sc_O2/600.)**-0.5 * 0.1/60./60                                                ## dm/s/l
        k_CO2_depth = k600_depth * (Sc_CO2/600.)**-0.5 * 0.1/60./60                                               ## dm/s/L
        K_CO2  = np.exp(-58.0931 + np.divide(90.5069*100,(T+273.15)) + 22.2940*np.log((T+273.15)/100))            ## mol/L/atm
    
        CO2_output = np.divide(K_CO2 * k_CO2_depth * pCO2_air + a * DOC * O2/(d+O2) * np.exp(c * pH -c*7.5), k_CO2_depth)            
        O2_output  = - np.divide(a*b * DOC * np.exp(c * pH -c*7.5) + k_O2_depth * (d-O2_air), 2*k_O2_depth) \
                + np.sqrt(np.divide(a*b * DOC * np.exp(c * pH -c*7.5) + k_O2_depth * (d-O2_air), 2*k_O2_depth)**2 + O2_air * d)   
    
        return np.concatenate((O2_output, CO2_output))

elif (decomposition_form == 'lin'):
    def function(x, a, b, d):
        
        Len = int(len(x[0])/2)
        DOC = x[0][:Len] ; O2 = x[1][:Len] ; pH = x[2][:Len]  ; T = x[3][:Len]                                    ## umol/L ; umol/L ; - ; °C
        
        k600_depth = 0.250751964                                                                                  ## cm/h/dm  
        
        O2_air   = 241.                                                                                           ## umol/L
        pCO2_air = 400.                                                                                           ## uatm
        Sc_O2  = 1568 - 86.04 * T + 2.142 * T**2 - 0.0216 * T**3
        Sc_CO2 = 1742 - 91.24 * T + 2.208 * T**2 - 0.0219 * T**3
        k_O2_depth  = k600_depth * (Sc_O2/600.)**-0.5 * 0.1/60./60                                                ## dm/s/l
        k_CO2_depth = k600_depth * (Sc_CO2/600.)**-0.5 * 0.1/60./60                                               ## dm/s/L
        K_CO2  = np.exp(-58.0931 + np.divide(90.5069*100,(T+273.15)) + 22.2940*np.log((T+273.15)/100))            ## mol/L/atm
        
        CO2_output = np.divide(K_CO2 * k_CO2_depth * pCO2_air + a * DOC * O2/(d+O2) * pH/7.5, k_CO2_depth) 
        O2_output  = - np.divide(a*b * DOC * pH/7.5 + k_O2_depth * (d-O2_air), 2*k_O2_depth) \
                + np.sqrt(np.divide(a*b * DOC * pH/7.5 + k_O2_depth * (d-O2_air), 2*k_O2_depth)**2 + O2_air * d)  
    
        return np.concatenate((O2_output, CO2_output))
    
elif (decomposition_form == 'exp_pH0'):
    def function(x, a, b, c, d, e):
        
        Len = int(len(x[0])/2)
        DOC = x[0][:Len] ; O2 = x[1][:Len] ; pH = x[2][:Len]  ; T = x[3][:Len]                                    ## umol/L ; umol/L ; - ; °C
        
        k600_depth = 0.250751964                                                                                  ## cm/h/dm  
        
        O2_air   = 241.                                                                                           ## umol/L
        pCO2_air = 400.                                                                                           ## uatm
        Sc_O2  = 1568 - 86.04 * T + 2.142 * T**2 - 0.0216 * T**3
        Sc_CO2 = 1742 - 91.24 * T + 2.208 * T**2 - 0.0219 * T**3
        k_O2_depth  = k600_depth * (Sc_O2/600.)**-0.5 * 0.1/60./60                                                ## dm/s/l
        k_CO2_depth = k600_depth * (Sc_CO2/600.)**-0.5 * 0.1/60./60                                               ## dm/s/L
        K_CO2  = np.exp(-58.0931 + np.divide(90.5069*100,(T+273.15)) + 22.2940*np.log((T+273.15)/100))            ## mol/L/atm
        
        CO2_output = np.divide(K_CO2 * k_CO2_depth * pCO2_air + a * DOC * O2/(d+O2) * np.exp(c * pH -c*e), k_CO2_depth)
        O2_output = - np.divide(a*b * DOC * np.exp(c * pH -c*e) + k_O2_depth * (d-O2_air), 2*k_O2_depth) \
                + np.sqrt(np.divide(a*b * DOC * np.exp(c * pH -c*e) + k_O2_depth * (d-O2_air), 2*k_O2_depth)**2 + O2_air * d)  
    
        return np.concatenate((O2_output, CO2_output))
    
elif (decomposition_form == 'pH_only'):
    def function(x, a, b, c):
        
        Len = int(len(x[0])/2)
        DOC = x[0][:Len] ; pH = x[2][:Len]  ; T = x[3][:Len]                                    ## umol/L ; umol/L ; - ; °C
        
        k600_depth = 0.250751964                                                                                  ## cm/h/dm  
        
        O2_air   = 241.                                                                                           ## umol/L
        pCO2_air = 400.                                                                                           ## uatm
        Sc_O2  = 1568 - 86.04 * T + 2.142 * T**2 - 0.0216 * T**3
        Sc_CO2 = 1742 - 91.24 * T + 2.208 * T**2 - 0.0219 * T**3
        k_O2_depth  = k600_depth * (Sc_O2/600.)**-0.5 * 0.1/60./60                                                ## dm/s/l
        k_CO2_depth = k600_depth * (Sc_CO2/600.)**-0.5 * 0.1/60./60                                               ## dm/s/L
        K_CO2  = np.exp(-58.0931 + np.divide(90.5069*100,(T+273.15)) + 22.2940*np.log((T+273.15)/100))            ## mol/L/atm
        
        CO2_output = np.divide(K_CO2 * k_CO2_depth * pCO2_air + a * DOC * np.exp(c * pH -c*7.5), k_CO2_depth)  
        O2_output = np.divide(k_O2_depth * O2_air - a * b * DOC * np.exp(c * pH -c*7.5), k_O2_depth)   
    
        return np.concatenate((O2_output, CO2_output))
    
elif (decomposition_form == 'O2_only'):
    def function(x, a, b, d):
        
        Len = int(len(x[0])/2)
        DOC = x[0][:Len] ; O2 = x[1][:Len] ; T = x[3][:Len]                                    ## umol/L ; umol/L ; - ; °C
        
        k600_depth = 0.250751964                                                                                  ## cm/h/dm  
        
        O2_air   = 241.                                                                                           ## umol/L
        pCO2_air = 400.                                                                                           ## uatm
        Sc_O2  = 1568 - 86.04 * T + 2.142 * T**2 - 0.0216 * T**3
        Sc_CO2 = 1742 - 91.24 * T + 2.208 * T**2 - 0.0219 * T**3
        k_O2_depth  = k600_depth * (Sc_O2/600.)**-0.5 * 0.1/60./60                                                ## dm/s/l
        k_CO2_depth = k600_depth * (Sc_CO2/600.)**-0.5 * 0.1/60./60                                               ## dm/s/L
        K_CO2  = np.exp(-58.0931 + np.divide(90.5069*100,(T+273.15)) + 22.2940*np.log((T+273.15)/100))            ## mol/L/atm
        
        CO2_output = np.divide(K_CO2 * k_CO2_depth * pCO2_air + a * DOC * O2/(d+O2), k_CO2_depth) 
        O2_output = - np.divide(a*b * DOC + k_O2_depth * (d-O2_air), 2*k_O2_depth) \
                + np.sqrt(np.divide(a*b * DOC + k_O2_depth * (d-O2_air), 2*k_O2_depth)**2 + O2_air * d)  
    
        return np.concatenate((O2_output, CO2_output))   
    
else: print('ERROR - choice of decomposition form does not exist')

#%%% Set estimated values for parameters
Rmax = 10**-6 ; Rmax_min = 0 ; Rmax_max = 1
b    = 0.8    ; b_min    = 0 ; b_max    = 2
L    = 0.6    ; L_min    = 0 ; L_max    = 1
Km   = 20     ; Km_min   = 0 ; Km_max   = 10**6
pH_0 = 7.5    ; pH_0_min = 3 ; pH_0_max = 10

if (decomposition_form == 'exp'): 
    p_0   = (Rmax, b, L, Km)
    p_min = [Rmax_min, b_min, L_min, Km_min]
    p_max = [Rmax_max, b_max, L_max, Km_max]
elif (decomposition_form == 'lin') or (decomposition_form == 'O2_only'):
    p_0   = (Rmax, b, Km)
    p_min = [Rmax_min, b_min, Km_min]
    p_max = [Rmax_max, b_max, Km_max]
elif (decomposition_form == 'exp_pH0'):
    p_0   = (Rmax, b, L, Km, pH_0)
    p_min = [Rmax_min, b_min, L_min, Km_min, pH_0_min]
    p_max = [Rmax_max, b_max, L_max, Km_max, pH_0_max]
elif (decomposition_form == 'pH_only'):
    p_0   = (Rmax, b, L)
    p_min = [Rmax_min, b_min, L_min]
    p_max = [Rmax_max, b_max, L_max]

#%%% Perform least-squares optimization
### least-squares optimization:
param, sigma = curve_fit(function, X_vals, np.concatenate((Data['O2(uM)'][mask],Data['CO2(uM)'][mask])),
                sigma=np.concatenate((Data['O2_err'][mask],Data['CO2_err'][mask])), p0 = p_0, bounds = (p_min,p_max))
param_std = np.sqrt(np.diag(sigma))

#%%% Print resulting values
print('Parameters derived for', decomposition_form, 'decomposition limitation:')
print('Rmax   = ', param[0]*10**6, 'pm', param_std[0]*10**6, 'umol/mol/s')
print('b      = ', param[1]*100, 'pm', param_std[1]*100, '%')
if (decomposition_form != 'lin') and (decomposition_form != 'O2_only'):
    print('lambda = ', param[2], 'pm', param_std[2])
if (decomposition_form == 'lin') or (decomposition_form == 'O2_only'):
    print('Km     = ', param[2], 'pm', param_std[2], 'uMol')
if (decomposition_form == 'exp') or (decomposition_form == 'exp_pH0'):
    print('Km     = ', param[3], 'pm', param_std[3], 'uMol')
if (decomposition_form == 'exp_pH0'):
    print('pH_0   = ', param[4], 'pm', param_std[4])

CO2 = Data['CO2(uM)'][mask] ; O2 = Data['O2(uM)'][mask]
print('')
print('CO2: R2   = ', stats.linregress(CO2, function(X_vals,*param)[len(CO2):])[2]**2)
print('CO2: P    = ', stats.linregress(CO2, function(X_vals,*param)[len(CO2):])[3])
print('CO2: RMSE = ', rmse(CO2, function(X_vals,*param)[len(CO2):]))
print('O2: R2    = ', stats.linregress(O2, function(X_vals,*param)[:len(O2)])[2]**2)
print('O2: P     = ', stats.linregress(O2, function(X_vals,*param)[:len(O2)])[3])
print('O2: RMSE  = ', rmse(O2, function(X_vals,*param)[:len(O2)]))